#!/usr/local/bin/perl -w

use strict;
use Getopt::Long;
use MIME::Parser;
use Fcntl;
use DB_File;

use vars qw($number_of_messages %token_occurrences);
use Bayespam::Process;

#
# bayes_process_email.pl
#
# This file is part of Bayespam. (v0.9.2)
#
# This program reads a corpus of spam email, a corpus of non-spam email,
#	and produces a corpus rating database describing the probability
#	of each token appearing in a spam email.  The file bayes_spam_check.pl
#	will use this corpus rating file to determine if a particular email is
#	spam or not.
#
# Copyright (C) 2002 Gary Arnold
#
# Bayespam is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Bayespam is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Bayespam; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

my %nonspam_occurrences = ();
my $number_of_nonspam_messages = 0;
my %spam_occurrences = ();
my $number_of_spam_messages = 0;
my %token_ratings = ();
my $mailcount;

my $maxsize = 100000;	 # ignore anything larger
my $recurse = 0;
my $case_sensitive = 0;
my @spammail = ();
my @goodmail = ();
my $help = 0;
my $remove = 0;
my $outfile = "";

my $parser = new MIME::Parser;

$parser->output_to_core(1);
$parser->extract_uuencode(1);

Getopt::Long::Configure ("bundling");
my $goodOptions = GetOptions ("maxsize|m=i" => \$maxsize,
							  "spam|s=s" => \@spammail,
							  "good|g=s" => \@goodmail,
							  "recurse|r" => \$recurse,
							  "delete|d" => \$remove,
							  "help|?|h" => \$help,
							  "out|o=s" => \$outfile,
							  "case|c"  => \$case_sensitive);

if ( ( $outfile eq "") || ( ( $#spammail < 0 ) && ( $#goodmail < 0 ) ) || $help || !$goodOptions )
{
	usage();
	exit 0;
}

# Open the database
tie( %token_ratings, "DB_File", $outfile, O_CREAT|O_RDWR, 0666, $DB_BTREE) ||
	die "Could not open or create database $outfile.";

# Disable STDOUT buffering
$| = 1;

# Read in saved settings in database
$case_sensitive = $token_ratings{ " case " } || $case_sensitive;
$number_of_spam_messages = $token_ratings{ " spam " } || 0;
$number_of_nonspam_messages = $token_ratings{ " nonspam " } || 0;

# Process the non-spam files
$mailcount = 0;
foreach (@goodmail) { ProcessDirectory( $_, \$number_of_nonspam_messages, \%nonspam_occurrences ); }
print "\nDone processing ".$mailcount." non-spam emails.\n" if $mailcount;

# Process the spam files
$mailcount = 0;
foreach (@spammail) { ProcessDirectory( $_, \$number_of_spam_messages, \%spam_occurrences ); }
print "\nDone processing ".$mailcount." spam emails.\n" if $mailcount;

# Store counts and case sensitivity, for incremental add and checking
$token_ratings{ " case " } = $case_sensitive;
$token_ratings{ " spam " } = $number_of_spam_messages || 0;
$token_ratings{ " nonspam " } = $number_of_nonspam_messages || 0;

my $count = 0;

# Go through the list of non-spam emails
foreach ( keys %nonspam_occurrences )
{
	if ( !exists $spam_occurrences{ $_ } )
	{
		# In this case, $badvalue would always be zero, so shortcut some math
		if ( ($spam_occurrences{ $_ } || 0) >= 5.0 )
		{
			$token_ratings{ $_ } = ($nonspam_occurrences{ $_ } || 0)." ".($spam_occurrences{ $_ } || 0)." ".0.01;
			$count++;
		}
	}
	else
	{
		# Good'uns count for double.
		my $goodvalue = 2.0 * $nonspam_occurrences{ $_ };
		my $badvalue = $spam_occurrences{ $_ };

		# Only consider tokens that occur often enough
		if ( ( $goodvalue + $badvalue ) >= 5.0 )
		{
			# Calculate the "spamminess" of this token
			$goodvalue /= $number_of_nonspam_messages;
			if ( $goodvalue > 1.0 ) { $goodvalue = 1.0; }

			$badvalue /= $number_of_spam_messages;
			if ( $badvalue > 1.0 ) { $badvalue = 1.0; }

			my $temp = $badvalue / ( $goodvalue + $badvalue );
			if ( $temp > 0.99 ) { $temp = 0.99; }
			if ( $temp < 0.01 ) { $temp = 0.01; }
			$token_ratings{ $_ } = ($nonspam_occurrences{ $_ } || 0)." ".($spam_occurrences{ $_ } || 0)." ".$temp;
			$count++;
		}
	}
}

# Do the same thing with the spams, but it's easier this time...think about it.
foreach ( keys %spam_occurrences )
{
	if ( !exists $nonspam_occurrences{ $_ } )
	{
		# In this case, $goodvalue would always be zero, so shortcut some math
		if ( ($spam_occurrences{ $_ } || 0) >= 5.0 )
		{
			$token_ratings{ $_ } = ($nonspam_occurrences{ $_ } || 0)." ".($spam_occurrences{ $_ } || 0)." ".0.99;
			$count++;
		}
	}
}
print "Done calculating corpus...shutting down\n";

untie( %token_ratings );


#################### SUBROUTINES ####################

# Given a directory name, go thru all the files in that directory,
#	treat each one as an email, and count how many times each token
#	appears in it.  A "token" is any group of alphanumerics, apostrophes,
#	dashes, and dollar signs -- words, in other words.
sub ProcessDirectory
{
	my $directory = shift(@_);
	local *number_of_messages = shift(@_);
	local *token_occurrences = shift(@_);
	my @files;
	my @token_list=();

	# Watch for STDIN special case
	if ( $directory eq "-" )
	{
		$number_of_messages += 1.0;
		ProcessFile( "-", \%token_occurrences );
		return;
	}

	if ( !opendir(DIR, $directory) )
	{
		print "Can't open directory $directory: $!\n";
		return;
	}
	@files = readdir(DIR);
	foreach ( @files )
	{
		if ( -d "$directory/$_" )
		{
			if ( /^\.\.?$/ )
			{
				next;
			}
			if ( $recurse )
			{
				my ( $R_number_of_messages, %R_token_occurrences ) = ProcessDirectory("$directory\/$_", $number_of_messages, %token_occurrences);
				if ( $remove ) { $number_of_messages -= $R_number_of_messages; } else { $number_of_messages += $R_number_of_messages; }

				foreach (keys %R_token_occurrences )
				{
					if ( $remove ) { $token_occurrences{ $_ } -= $R_token_occurrences{ $_ }; } else { $token_occurrences{ $_ } += $R_token_occurrences{ $_ }; }
				}
			}
		}
		elsif ( -f "$directory/$_" )
		{
			my $fsize = (stat("$directory/$_"))[7];
			if ( $fsize > $maxsize )
			{
				next;
			}

			if ( $remove ) { $number_of_messages -= 1.0; } else { $number_of_messages += 1.0; }
			ProcessFile( "$directory/$_", \%token_occurrences );
		}
	}

	closedir(DIR);
}

# Given a file name, count how many times each token
#	appears in it.  A "token" is any group of alphanumerics, apostrophes,
#	dashes, and dollar signs -- words, in other words.
sub ProcessFile
{
	my $file = shift(@_);
	local *token_occurrences = shift(@_);

	$mailcount++;
	print '.' if ( $mailcount % 100 ) == 0;

	if ( ( $file eq "-" ) || open(F, "< $file") )
	{
		my $data;
		$data = $parser->parse(\*F) if $file ne "-";
		$data = $parser->parse(\*STDIN) if $file eq "-";

		my $message = Bayespam::Process::ProcessMimeMessage( $data );

		# Rip out HTML comments
		$message =~ s/\<\!\-.*\-\>//gs;
		$message = lc $message if not $case_sensitive;

		foreach (split( /[^\w'$-]+/, $message ) )
		{
			# Ignore all-numeric tokens
			if ( !( $_ =~ /^[0-9\-]+$/g ) ) { if (length($_) > 0){$token_occurrences{ $_ } += 1.0;} }
		}
	}
}

# Print program usage
sub usage
{
    print STDERR << "EOF";

Usage: $0 [-hdrs] [-m max] [-g path] [-b path] -o file
Add or remove an email corpus to a Bayespam rating corpus.

 -h, --help          : this (help) message
 -d, --delete        : remove processed emails from database
 -r, --recurse       : recurse subsirectories
 -c, --case          : case sensitive tokens
 -m, --maxsize=SIZE  : ignore any file larger than SIZE bytes (default 100000)
 -g, --good=PATH     : path to good email file or directory
 -s, --spam=PATH     : path to spam email file or directory
 -o, --out=FILE      : output rating file

example: $0 --case --good ./Maildir --spam ./Maildir/.Spam -o bayes_rating.dat

EOF
}
